'''
Module containing utilities for the package
'''
# Import Packages
import numpy as np
import xarray as xr
import datetime
from .Gridder import Gridder

# Define functions and classes
def print_time(message):
    '''Add the current time to the end of a message'''
    # Get current time
    current_time = datetime.datetime.today().strftime('%H:%M:%S')
    # Add time to message
    whole_message = f'{message}: {current_time}'
    # Print out the message
    print(whole_message)

def find_nth(haystack: str, needle: str, n: int) -> int:
    start = haystack.find(needle)
    while start >= 0 and n > 1:
        start = haystack.find(needle, start+len(needle))
        n -= 1
    return start

def invert_dict(dict:dict) -> dict:
    return {value:key for key,value in dict.items()}

def add_gridded_data(ds_mission:xr.Dataset) -> xr.Dataset:
    '''Create gridder object and create the gridded dataset'''
    print_time('Adding Gridded Data')
    gridder = Gridder(ds_mission=ds_mission)
    gridder.create_gridded_dataset()
    ds_mission.update(gridder.ds_gridded)
    print_time('Finished Adding Gridded Data')
    return ds_mission

def get_polygon_coords(ds_mission:xr.Dataset) -> str:
    '''Get the polygon coords for the dataset global attributes'''
    lat_max = np.nanmax(ds_mission.latitude[np.where(ds_mission.latitude.values<29.5)].values)
    lat_min = np.nanmin(ds_mission.latitude[np.where(ds_mission.latitude.values<29.5)].values)
    lon_max = np.nanmax(ds_mission.longitude.values)
    lon_min = np.nanmin(ds_mission.longitude.values)
    polygon_1 = str(lat_max)+' '+str(ds_mission.longitude[np.where(ds_mission.latitude==lat_max)[0][0]].values) # northmost
    polygon_2 = str(ds_mission.latitude[np.where(ds_mission.longitude==lon_max)[0][0]].values)+' '+str(lon_max) # eastmost
    polygon_3 = str(lat_min)+' '+str(ds_mission.longitude[np.where(ds_mission.latitude==lat_min)[0][0]].values) # southmost
    polygon_4 = str(ds_mission.latitude[np.where(ds_mission.longitude==lon_min)[0][0]].values)+' '+str(lon_min) # westmost
    polygon_5 = polygon_1
    return 'POLYGON (('+polygon_1+' '+polygon_2+' '+polygon_3+' '+polygon_4+' '+polygon_5+'))'



import pandas as pd
import xarray as xr
from pathlib import Path
from attrs import define,field
import datetime


from glider_ingest.utils import find_nth,invert_dict

@define
class MissionData:
    # Required Variables
    memory_card_copy_loc:Path
    working_dir:Path
    mission_num:str

    # Optional Variables
    glider_id:str = field(default=None)
    nc_filename:str = field(default=None)
    output_nc_path:Path = field(default=None)
    mission_start_date:str|None = field(default=None)
    mission_end_date:str|None = field(default=None)
    mission_year:str = field(default=None)
    mission_title:str = field(default=None)
    glider_name:str = field(default=None)
    glider_ids:dict = field(default={'199':'Dora','307':'Reveille','308':'Howdy','540':'Stommel','541':'Sverdrup','1148':'unit_1148'})
    wmo_ids:dict = field(default={'199':'unknown','307':'4801938','308':'4801915','540':'4801916','541':'4801924','1148':'4801915'})
    wmo_id:str = field(default=None)

    # Post init variables
    fli_files_loc:Path = field(init=False)
    fli_cache_loc:Path = field(init=False)
    sci_files_loc:Path = field(init=False)
    sci_cache_loc:Path = field(init=False)

    # Data Ingest Variables
    df_fli:pd.DataFrame = field(init=False)
    ds_fli:xr.Dataset = field(init=False)

    df_sci:pd.DataFrame = field(init=False)
    ds_sci:xr.Dataset = field(init=False)

    ds_mission:xr.Dataset = field(init=False)
    
    
    def __attrs_post_init__(self):
        self.get_file_locs()

    def setup(self):
        self.get_mission_date_range()
        self.get_mission_year_and_glider()
        self.get_wmo_id()
        self.get_mission_title()
        self.get_nc_filename()
        self.get_output_nc_path()
        

    def get_file_locs(self):
        def set_path_with_case(base_path: Path, *parts: str) -> Path:
            path = base_path.joinpath(*parts)
            if path.exists():
                return path
            return base_path.joinpath(*parts[:-1], parts[-1].upper())

        self.fli_files_loc = set_path_with_case(self.memory_card_copy_loc, 'Flight_card', 'logs')
        self.fli_cache_loc = set_path_with_case(self.memory_card_copy_loc, 'Flight_card', 'state', 'cache')
        self.sci_files_loc = set_path_with_case(self.memory_card_copy_loc, 'Science_card', 'logs')
        self.sci_cache_loc = set_path_with_case(self.memory_card_copy_loc, 'Science_card', 'state', 'cache')

    def get_mission_date_range(self):
        if self.mission_end_date is None:
            self.mission_end_date = str(datetime.datetime.today().date()+datetime.timedelta(days=365))
        if self.mission_start_date is None:
            self.mission_start_date = '2010-01-01'

    def get_mission_year_and_glider(self):
        file = self._get_sample_file()
        name = self._extract_full_filename(file)
        self._parse_mission_year(name)
        self._parse_and_validate_glider_name(name)

    def _get_sample_file(self):
        files = self.get_files(files_loc=self.sci_files_loc, extension='ebd')
        return files[10]

    def _extract_full_filename(self, file):
        with open(file, errors="ignore") as fp:
            for line in fp:
                if 'full_filename' in line.strip():
                    return line.replace('full_filename:', '').strip()
        return None

    def _parse_mission_year(self, name):
        self.mission_year = name[name.find('-')+1:find_nth(name,'-',2)].strip()

    def _parse_and_validate_glider_name(self, name):
        # Extract glider name from the input string
        glider_name = name.split('-')[0].replace('unit_', '').strip()
        
        # Create bidirectional mapping
        inverted_glider_ids = {v: k for k, v in self.glider_ids.items()}
        
        # If glider_name is an ID, get the name
        if glider_name in self.glider_ids:
            self.glider_name = self.glider_ids[glider_name]
            self.glider_id = glider_name
            return
            
        # If glider_name is a name, get the ID
        if glider_name in inverted_glider_ids:
            self.glider_name = glider_name
            self.glider_id = inverted_glider_ids[glider_name]
            return
            
        # If we get here, the glider name/id wasn't found
        valid_options = list(self.glider_ids.keys()) + list(self.glider_ids.values())
        raise ValueError(f'Invalid glider identifier: {glider_name}. Must be one of: {valid_options}')


    def get_mission_title(self):
        if self.mission_title is None:
            self.mission_title = f'Mission {self.mission_num}'

    def get_nc_filename(self):
        if self.nc_filename is None:
            self.nc_filename = f'M{self.mission_num}_{self.mission_year}_{self.glider_id}.nc'

    def get_output_nc_path(self):
        if self.output_nc_path is None:
            output_nc_loc = self.working_dir.joinpath(self.mission_title)
            output_nc_loc.mkdir(exist_ok=True,parents=True)
            self.output_nc_path = output_nc_loc.joinpath(self.nc_filename)
        # If self.output_nc_path is a string
        if isinstance(self.output_nc_path,str):
            self.output_nc_path = Path(self.output_nc_path)
        # Ensure self.output_nc_path is a pathlib.Path object
        if isinstance(self.output_nc_path,Path):
            # If the provided output_nc_path does not specify the filename
            if not self.output_nc_path.is_file():
                self.output_nc_path.joinpath(self.nc_filename)

    def get_wmo_id(self):
        '''Get glider wmo id from glider id and dict of wmo ids'''
        if self.wmo_id is None:
            self.wmo_id = self.wmo_ids[self.glider_id]

    def get_files(self,files_loc:Path,extension:str):
        '''Get files to process from files_loc'''

        if files_loc.exists():
            try:
                files = list(files_loc.rglob(f'*.{extension.lower()}'))
                files = [str(file) for file in files]
                if len(files) ==0 :
                    raise ValueError(f'No Files found at {files_loc.resolve()}')
            except ValueError:
                files = list(files_loc.rglob(f'*.{extension.upper()}'))
                files = [str(file) for file in files]
                if len(files) ==0 :
                    raise ValueError(f'No Files found at {files_loc.resolve()}')
            return files
        else: 
            raise ValueError(f'Path not found: {files_loc.resolve()}')
        


from pathlib import Path

from glider_ingest import MissionData, MissionProcessor

memory_card_copy_loc = Path('path/to/memory/card/copy')
# Where you want the netcdf to be saved to
working_dir = Path('path/to/working/dir').resolve()
mission_num = '46'

# Initalize the mission_data container
mission_data = MissionData(memory_card_copy_loc=memory_card_copy_loc,
                         working_dir=working_dir,
                         mission_num=mission_num)
# Pass the mission_data container to the MissionProcessor class
# call save_mission_dataset to generate and save the mission dataset
MissionProcessor(mission_data=mission_data).save_mission_dataset()



import numpy as np
import pandas as pd
import xarray as xr
import dbdreader
from attrs import define

from glider_ingest.MissionData import MissionData
from glider_ingest.utils import print_time


@define
class FlightProcessor:

    mission_data:MissionData
  
    def load_flight(self):
        files = self.mission_data.get_files(files_loc=self.mission_data.fli_files_loc,extension='dbd')
        dbd = dbdreader.MultiDBD(files,cacheDir=self.mission_data.fli_cache_loc)
        data = dbd.get_sync('m_lat', 'm_lon', 'm_pressure','m_water_depth')

        self.mission_data.df_fli = pd.DataFrame(data).T
        self.mission_data.df_fli.columns = ['m_present_time', 'm_lat', 'm_lon', 'm_pressure','m_water_depth']
        self.mission_data.df_fli['m_present_time'] = pd.to_datetime(self.mission_data.df_fli['m_present_time'],unit='s')

        # Remove any data with erroneous dates
        self.mission_data.df_fli = self.mission_data.df_fli.loc[(self.mission_data.df_fli['m_present_time'] > self.mission_data.mission_start_date) & (self.mission_data.df_fli['m_present_time'] < self.mission_data.mission_end_date)]
        # Convert pressure from db to dbar
        self.mission_data.df_fli['m_pressure'] *= 10

        # Rename columns for clarity
        self.mission_data.df_fli.rename(columns={'m_lat': 'm_latitude', 'm_lon': 'm_longitude'}, inplace=True)
        self.mission_data.df_fli = self.mission_data.df_fli.dropna()
        dbd.close()

    def convert_fli_df_to_ds(self) -> xr.Dataset:
        '''Convert the flight dataframe to dataset'''
        self.mission_data.ds_fli = xr.Dataset.from_dataframe(self.mission_data.df_fli)


    def add_flight_attrs(self) -> xr.Dataset:
        '''Add attributes to the flight dataset'''
        self.mission_data.ds_fli['m_pressure'].attrs = {'accuracy': 0.01,
        'ancillary_variables': ' ',
        'axis': 'Z',
        'bytes': 4,
        'comment': 'Alias for m_pressure, multiplied by 10 to convert from bar to dbar',
        'long_name': 'GPS Pressure',
        'observation_type': 'measured',
        'platform': 'platform',
        'positive': 'down',
        'precision': 0.01,
        'reference_datum': 'sea-surface',
        'resolution': 0.01,
        'source_sensor': 'sci_water_pressure',
        'standard_name': 'sea_water_pressure',
        'units': 'bar',
        'valid_max': 2000.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        self.mission_data.ds_fli['m_water_depth'].attrs = {'accuracy': 0.01,
        'ancillary_variables': ' ',
        'axis': 'Z',
        'bytes': 4,
        'comment': 'Alias for m_depth',
        'long_name': 'GPS Depth',
        'observation_type': 'calculated',
        'platform': 'platform',
        'positive': 'down',
        'precision': 0.01,
        'reference_datum': 'sea-surface',
        'resolution': 0.01,
        'source_sensor': 'm_depth',
        'standard_name': 'sea_water_depth',
        'units': 'meters',
        'valid_max': 2000.0,
        'valid_min': 0.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        self.mission_data.ds_fli['m_latitude'].attrs = {'ancillary_variables': ' ',
        'axis': 'Y',
        'bytes': 8,
        'comment': 'm_gps_lat converted to decimal degrees and interpolated',
        'coordinate_reference_frame': 'urn:ogc:crs:EPSG::4326',
        'long_name': 'Latitude',
        'observation_type': 'calculated',
        'platform': 'platform',
        'precision': 5,
        'reference': 'WGS84',
        'source_sensor': 'm_gps_lat',
        'standard_name': 'latitude',
        'units': 'degree_north',
        'valid_max': 90.0,
        'valid_min': -90.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}
        self.mission_data.ds_fli['m_longitude'].attrs = {'ancillary_variables': ' ',
        'axis': 'X',
        'bytes': 8,
        'comment': 'm_gps_lon converted to decimal degrees and interpolated',
        'coordinate_reference_frame': 'urn:ogc:crs:EPSG::4326',
        'long_name': 'Longitude',
        'observation_type': 'calculated',
        'platform': 'platform',
        'precision': 5,
        'reference': 'WGS84',
        'source_sensor': 'm_gps_lon',
        'standard_name': 'longitude',
        'units': 'degree_east',
        'valid_max': 180.0,
        'valid_min': -180.0,
        'update_time': pd.Timestamp.now().strftime(format='%Y-%m-%d %H:%M:%S')}


    def format_flight_ds(self) -> xr.Dataset:
        '''Format the flight dataset by sorting and renaming variables'''
        self.mission_data.ds_fli['index'] = np.sort(self.mission_data.ds_fli['m_present_time'].values.astype('datetime64[ns]'))
        self.mission_data.ds_fli = self.mission_data.ds_fli.drop_vars('m_present_time')
        self.mission_data.ds_fli = self.mission_data.ds_fli.rename({'index': 'm_time','m_pressure':'m_pressure','m_water_depth':'depth','m_latitude':'latitude','m_longitude':'longitude'})


    def process_flight_data(self) -> xr.Dataset:
        '''Perform all processing of flight data from dbd to pandas dataframe to xarray dataset'''
        # Process Flight Data
        print_time('Processing Flight Data')
        self.load_flight()
        self.convert_fli_df_to_ds()
        self.add_flight_attrs()
        self.format_flight_ds()
        print_time('Finised Processing Flight Data')
